
"""Model for fitting torch models.
"""
import os
import time
import random
from collections import OrderedDict, defaultdict
import warnings
import contextlib
import numpy as np  # shoul be remved
import torch
import torchtuples.callbacks as cb
from torchtuples.optim import AdamW, OptimWrap
from torchtuples.tupletree import tuplefy, TupleTree, make_dataloader
from torchtuples.utils import make_name_hash


class Model(object):
    """Train torch models using dataloaders, tensors or np.arrays.
    
    Arguments:
        net {torch.nn.Module} -- A torch module.
    
    Keyword Arguments:
        loss {function} -- Set function that is used for training 
            (e.g. binary_cross_entropy for torch) (default: {None})
        optimizer {Optimizer} -- A torch optimizer or similar. Preferrably use torchtuples.optim instead of
            torch.optim, as this allows for reinitialization, etc. If 'None' set to torchtuples.optim.AdamW.
            (default: {None})
        device {str, int, torch.device} -- Device to compute on. (default: {None})
            Preferrably pass a torch.device object.
            If 'None': use default gpu if avaiable, else use cpu.
            If 'int': used that gpu: torch.device('cuda:<device>').
            If 'string': string is passed to torch.device('string').

    Example simple model:
    ---------------------
    from torchtuples import Model
    import torch
    from torch import nn
    from torch.nn import functional as F

    n_rows, n_covs = 1000, 4
    x = torch.randn((n_rows, n_covs))
    y = 2 * x.sum(1, keepdim=True) + 4  # y = 2 * x + 4

    net = nn.Sequential(nn.Linear(n_covs, 10), nn.ReLU(), nn.Linear(10, 1))
    loss = F.mse_loss
    model = Model(net, loss)
    log = model.fit(x, y, batch_size=32, epochs=30)
    log.plot()
    """
    def __init__(self, net, loss=None, optimizer=None, device=None):
        self.net = net
        if type(self.net) is str:
            self.load_net(self.net)
        self.loss = loss
        self.optimizer = optimizer if optimizer is not None else AdamW
        self.device = self._device_from__init__(device)
        self.net.to(self.device)
        # self.net_predict = net_predict if net_predict else self.net
        # self.net_predict.to(self.device)
        if not hasattr(self, 'make_dataloader_predict'):
            self.make_dataloader_predict = self.make_dataloader
        self._init_train_log()

    def _init_train_log(self):
        self.log = cb.TrainingLogger()
        self.train_metrics = cb._MonitorFitMetricsTrainData()
        self.val_metrics = cb.MonitorFitMetrics()
        self.log.monitors = OrderedDict(train_=self.train_metrics, val_=self.val_metrics)
        self.callbacks = None
    
    @staticmethod
    def _device_from__init__(device):
        if device is None:
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        elif type(device) is str:
            device = torch.device(device)
        elif type(device) is int:
            device = torch.device(f"cuda:{device}")
        if type(device) is not torch.device:
            raise ValueError("Argument 'device' needs to be None, string, int or torch.device object.")
        return device
    
    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer):
        self._optimizer = optimizer
        if callable(self._optimizer):
            self._optimizer = self._optimizer(params=self.net.parameters())
        if not isinstance(self._optimizer, OptimWrap):
            self._optimizer = OptimWrap(self._optimizer)

    @staticmethod
    def make_dataloader(data, batch_size, shuffle, num_workers=0, **kwargs):
        """Function for creating a dataloader from tensors or arrays.
        It is natural to rewrite this method in inherited classes.

        self.make_dataloader_predict will be set to this method if not implemented
        separatelly.

        This simply calls tupletree.make_dataloader, but is included to make
        inheritance simpler. 

        Arguments:
            data {tuple, np.array, tensor} -- Data in dataloader e.g. (x, y)
            batch_size {int} -- Batch size used in dataloader
            shuffle {bool} -- If order should be suffled

        Keyword Arguments:
            num_workers {int} -- Number of workers in dataloader (default: {0})
            to_tensor {bool} -- Ensure that we use tensors (default: {True})
            **kwargs -- Passed to make_dataloader.

        Returns:
            DataLoaderSlice -- A dataloader object like the torch DataLoader
        """
        dataloader = make_dataloader(data, batch_size, shuffle, num_workers, **kwargs)
        return dataloader
    
    def _setup_train_info(self, dataloader):
        self.fit_info = {'batches_per_epoch': len(dataloader)}
        def _tuple_info(tuple_):
            tuple_ = tuplefy(tuple_)
            return {'levels': tuple_.to_levels(), 'shapes': tuple_.shapes().apply(lambda x: x[1:])}
        # input, target = dataloader.dataset[0]
        data = _get_element_in_dataloader(dataloader)
        if data is not None:
            input, target = data
            self.fit_info['input'] = _tuple_info(input)
    
    def _to_device(self, data):
        """Move data to self.device.
        
        Arguments:
            data {tensor or tuple} -- Data
        
        Returns:
            tensor or tuple -- Data moved to device.
        """
        if data is None:
            return tuplefy(data)
        return tuplefy(data).to_device(self.device)

    def compute_metrics(self, input, target, metrics):
        """Function for computing loss.
        Is rather general, but can be reimpliemented by sub classes.
        
        Arguments:
            input {tensor or tuple} -- This should be passed to self.net.
            target {tensor or tuple} -- This is the targets that should be used in self.loss.
            metrics (dictionary with funcs) -- Metrics that should be computed.

        Returns:
            tensor -- Results of self.loss()
        """
        if (self.loss is None) and (self.loss in metrics.values()):
            raise RuntimeError(f"Need to specify a loss (self.loss). It's currently None")
        input = self._to_device(input)
        target = self._to_device(target)
        out = self.net(*input)
        out = tuplefy(out)
        res = {name: metric(*out, *target) for name, metric in metrics.items()}
        return res

    def _setup_metrics(self, metrics):
        all_metrics = {'loss': self.loss}
        if metrics is not None:
            if not hasattr(metrics, 'items'):
                if not hasattr(metrics, '__iter__'):
                    metrics = [metrics]
                metrics = {met.__name__: met for met in metrics}
            if 'loss' in metrics:
                raise ValueError("The 'loss' keyword is reserved for the loss function.")
            all_metrics.update(metrics)
        return all_metrics

    def fit_dataloader(self, dataloader, epochs=1, callbacks=None, verbose=True, metrics=None,
                       val_dataloader=None):
        """Fit a dataloader object.
        See 'fit' for tensors and np.arrays.
        
        Arguments:
            dataloader {dataloader} -- A dataloader that gives (input, target).
        
        Keyword Arguments:
            epochs {int} -- Number of epochs (default: {1})
            callbacks {list} -- list of callbacks (default: {None})
            verbose {bool} -- Print progress (default: {True})
        
        Returns:
            TrainingLogger -- Training log
        """
        self._setup_train_info(dataloader)
        self.metrics = self._setup_metrics(metrics)
        self.log.verbose = verbose
        self.val_metrics.dataloader = val_dataloader
        if callbacks is None:
            callbacks = []
        self.callbacks = cb.TrainingCallbackHandler(self.optimizer, self.train_metrics, self.log,
                                                    self.val_metrics, callbacks)
        self.callbacks.give_model(self)

        stop = self.callbacks.on_fit_start()
        for _ in range(epochs):
            if stop: break
            stop = self.callbacks.on_epoch_start()
            if stop: break
            for input, target in dataloader:
                stop = self.callbacks.on_batch_start()
                if stop: break
                self.optimizer.zero_grad()
                self.batch_metrics = self.compute_metrics(input, target, self.metrics)
                self.batch_loss = self.batch_metrics['loss']
                self.batch_loss.backward()
                stop = self.callbacks.before_step()
                if stop: break
                self.optimizer.step()
                stop = self.callbacks.on_batch_end()
                if stop: break
            else:
                stop = self.callbacks.on_epoch_end()
        self.callbacks.on_fit_end()
        return self.log

    def fit(self, input, target, batch_size=256, epochs=1, callbacks=None, verbose=True,
            num_workers=0, shuffle=True, metrics=None, val_data=None, val_batch_size=8224,
            **kwargs):
        """Fit  model with inputs and targets.
        
        Arguments:
            input {np.array, tensor or tuple} -- Input (x) passed to net.
            target {np.array, tensor or tuple} -- Target (y) passed to loss function.
        
        Keyword Arguments:
            batch_size {int} -- Elemets in each batch (default: {256})
            epochs {int} -- Number of epochs (default: {1})
            callbacks {list} -- list of callbacks (default: {None})
            verbose {bool} -- Print progress (default: {True})
            num_workers {int} -- Number of workers used in the dataloader (default: {0})
            shuffle {bool} -- If we should shuffle the order of the dataset (default: {True})
            **kwargs -- Passed to the 'make_dataloader' method. Set e.g. `torch_ds_dl to use
                the TensorDataset and DataLoader provided by torch instead of the torchtuples
                implementations.
    
        Returns:
            TrainingLogger -- Training log
        """
        dataloader = self.make_dataloader((input, target), batch_size, shuffle, num_workers, **kwargs)
        val_dataloader = val_data
        if type(val_data) in (list, tuple, TupleTree):
            val_dataloader = self.make_dataloader(val_data, val_batch_size, shuffle=False,
                                                  num_workers=num_workers, **kwargs)
        log = self.fit_dataloader(dataloader, epochs, callbacks, verbose, metrics, val_dataloader)
        return log

    @contextlib.contextmanager
    def _lr_finder(self, lr_min, lr_max, lr_range, n_steps, tolerance, verbose):
        lr_lower, lr_upper = lr_range
        path = make_name_hash('lr_finder_checkpoint')
        self.save_model_weights(path)
        try:
            self.optimizer.drop_scheduler()
            lr_finder = cb.LRFinder(lr_lower, lr_upper, n_steps, tolerance)
            yield lr_finder
        except Exception as e:
            self.load_model_weights(path)
            os.remove(path)
            self.optimizer = self.optimizer.reinitialize()
            self._init_train_log()
            raise e
        self.load_model_weights(path)
        os.remove(path)
        lr = lr_finder.get_best_lr(lr_min, lr_max)
        self.optimizer = self.optimizer.reinitialize(lr=lr)
        self._init_train_log()

    def lr_finder(self, input, target, batch_size=256, lr_min=1e-4, lr_max=1., lr_range=(1e-7, 10.),
                  n_steps=100, tolerance=np.inf, callbacks=None, verbose=False, num_workers=0,
                  shuffle=True, **kwargs):
        with self._lr_finder(lr_min, lr_max, lr_range, n_steps, tolerance, verbose) as lr_finder:
            if callbacks is None:
                callbacks = []
            callbacks.append(lr_finder)
            epochs = n_steps
            self.fit(input, target, batch_size, epochs, callbacks, verbose, num_workers,
                        shuffle, **kwargs)
        return lr_finder

    def lr_finder_dataloader(self, dataloader, lr_min=1e-4, lr_max=1., lr_range=(1e-7, 10.),
                             n_steps=100, tolerance=np.inf, callbacks=None, verbose=False):
        with self._lr_finder(lr_min, lr_max, lr_range, n_steps, tolerance, verbose) as lr_finder:
            if callbacks is None:
                callbacks = []
            callbacks.append(lr_finder)
            epochs = n_steps
            self.fit_dataloader(dataloader, epochs, callbacks, verbose)
        return lr_finder

    def score_in_batches(self, data, score_func=None, batch_size=8224, eval_=True, mean=True,
                         num_workers=0, shuffle=False, make_dataloader=None, numpy=True, **kwargs):
        """Used to score a dataset in batches.
        If score_func is None, this use the loss function.
        If make_dataloader is None, we use self.make_dataloader_predict, unless score_func is also
        None, in which we use self.make_dataloader.
        
        Arguments:
            data {np.array, tensor, tuple, dataloader} -- Data in the form a datloader, or arrarys/tensors.
        
        Keyword Arguments:
            score_func {func} -- Function used for scoreing. If None, we use self.loss. (default: {None})
            batch_size {int} -- Batch size (default: {8224})
            eval_ {bool} -- Eval mode of the net. (default: {True})
            mean {bool} -- If True, we return the mean. (default: {True})
            num_workers {int} -- Number of workers for the dataloader. (default: {0})
            shuffle {bool} -- If the data should be shuffled (default: {False})
            make_dataloader {func} -- Function for making a dataloder.
                If None, we use make_dataloader_predict as long as score_func is not None. (default: {None})
            **kwargs -- Are passed to make_dataloader function.
        
        Returns:
            np.array -- Scores
        """
        if make_dataloader is None:
            if score_func is None:
                make_dataloader = self.make_dataloader
            else:
                make_dataloader = self.make_dataloader_predict
        if data.__class__ in (list, tuple, TupleTree):
            data = make_dataloader(data, batch_size, shuffle, num_workers, **kwargs)
        scores = self.score_in_batches_dataloader(data, score_func, eval_, mean, numpy)
        return scores
    
    def score_in_batches_dataloader(self, dataloader, score_func=None, eval_=True, mean=True, numpy=True):
        '''Score a dataset in batches.

        Parameters:
            dataloader: Dataloader:
            score_func: Function of (self, data) that returns a measure.
                If None, we get training loss.
            eval_: If net should be in eval mode.
            mean: If return mean or list with scores.
        '''
        if eval_:
            self.net.eval()
        batch_scores = []
        with torch.no_grad():
            for input, target in dataloader:
                if score_func is None:
                    # score = self.compute_loss(input, target)
                    score = self.compute_metrics(input, target, self.metrics)
                else:
                    warnings.warn(f"score_func {score_func} probably doesn't work... Not implemented")
                    score = score_func(self, input, target)
                batch_scores.append(score)
        if eval_:
            self.net.train()
        if type(batch_scores[0]) is dict:
            scores = defaultdict(list)
            for bs in batch_scores:
                for name, score in bs.items():
                    scores[name].append(score)
            scores = {name: torch.tensor(score) for name, score in scores.items()}
            if mean:
                scores = {name: score.mean() for name, score in scores.items()}
            if numpy:
                # scores = {name: tuplefy(score).to_numpy()[0] for name, score in scores.items()}
                scores = {name: score.item() for name, score in scores.items()}
            return scores
        if mean:
            batch_scores = [score.item() for score in batch_scores]
            return np.mean(batch_scores)
        return batch_scores

    def predict(self, input, batch_size=8224, numpy=None, eval_=True,
                grads=False, to_cpu=False, num_workers=0, **kwargs):
        """Get predictions from 'input'.
        
        Arguments:
            input {tuple, np.ndarra, or torch.tensor} -- Input to net.
        
        Keyword Arguments:
            batch_size {int} -- Batch size (default: {8224})
            numpy {bool} -- 'False' gives tensor, 'True' gives numpy, and None give same as input
                (default: {None})
            eval_ {bool} -- If 'True', use 'eval' modede on net. (default: {True})
            grads {bool} -- If gradients should be computed (default: {False})
            to_cpu {bool} -- For larger data sets we need to move the results to cpu
                (default: {False})
            num_workers {int} -- Number of workes in created dataloader (default: {0})
            **kwargs -- Passed to make_dataloader.
        
        Returns:
            [TupleTree, np.ndarray or tensor] -- Predictions
        """
        if numpy is None:
            numpy = tuplefy(input).type() is not torch.Tensor
        dataloader = self.make_dataloader_predict(input, batch_size, shuffle=False,
                                                  num_workers=num_workers, **kwargs)
        preds = self.predict_dataloader(dataloader, numpy, eval_, grads, to_cpu)
        return preds

    def predict_dataloader(self, dataloader, numpy=True, eval_=True, grads=False, to_cpu=False):
        """Get predictions from dataloader.
        
        Arguments:
            dataloader {DataLoader} -- Dataloader with inputs to net.
        
        Keyword Arguments:
            numpy {bool} -- If 'False', tensor is returned (default: {True})
            eval_ {bool} -- If 'True', use 'eval' modede on net. (default: {True})
            grads {bool} -- If gradients should be computed (default: {False})
            to_cpu {bool} -- For larger data sets we need to move the results to cpu
                (default: {False})
        
        Returns:
            [TupleTree, np.ndarray or tensor] -- Predictions
        """
        if hasattr(self, 'fit_info') and (self.make_dataloader is self.make_dataloader_predict):
            # input = tuplefy(dataloader.dataset[0])
            data = _get_element_in_dataloader(dataloader)
            if data is not None:
                input = tuplefy(data)
                input_train = self.fit_info['input']
                if input.to_levels() != input_train['levels']:
                    raise RuntimeError("""The input from the dataloader is different from
                    the 'input' during trainig. Make sure to remove 'target' from dataloader.
                    Can be done with 'torchtuples.data.dataloader_input_only'.""")
                if input.shapes().apply(lambda x: x[1:]) != input_train['shapes']:
                    raise RuntimeError("""The input from the dataloader is different from
                    the 'input' during trainig. The shapes are different.""")

        if not eval_:
            warnings.warn("We still don't shuffle the data here... event though 'eval_' is True.")
        if eval_:
            self.net.eval()
        with torch.set_grad_enabled(grads):
            preds = []
            for input in dataloader:
                input = tuplefy(input).to_device(self.device)
                preds_batch = tuplefy(self.net(*input))
                if numpy or to_cpu:
                    preds_batch = preds_batch.to_device('cpu')
                preds.append(preds_batch)
        if eval_:
            self.net.train()
        preds = tuplefy(preds).cat()
        if numpy:
            preds = preds.to_numpy()
        if len(preds) == 1:
            preds = preds[0]
        return preds

    def save_model_weights(self, path, **kwargs):
        '''Save the model weights.

        Parameters:
            path: The filepath of the model.
            **kwargs: Arguments passed to torch.save method.
        '''
        return torch.save(self.net.state_dict(), path, **kwargs)

    def load_model_weights(self, path, **kwargs):
        '''Load model weights.

        Parameters:
            path: The filepath of the model.
            **kwargs: Arguments passed to torch.load method.
        '''
        self.net.load_state_dict(torch.load(path, **kwargs))

    def save_net(self, path, **kwargs):
        """Save self.net to file (e.g. net.pt).
        
        Arguments:
            path {str} -- Path to file.
            **kwargs are passed to torch.save
        
        Returns:
            None
        """
        return torch.save(self.net, path, **kwargs)

    def load_net(self, path, **kwargs):
        """Load net from file (e.g. net.pt), and set as self.net
        
        Arguments:
            path {str} -- Path to file.
            **kwargs are passed to torch.load
        
        Returns:
            torch.nn.Module -- self.net
        """
        self.net = torch.load(path, **kwargs)
        if hasattr(self, 'device'):
            self.net.to(self.device)
        return self.net

def _get_element_in_dataloader(dataloader):
    dataset = dataloader.dataset
    try:
        return dataset[:2]
    except:
        pass
    try:
        return dataset[[0, 1]]
    except:
        pass
    try:
        return dataloader.collate_fn([dataset[0], dataset[1]])
    except:
        pass
    return None



import time
import random

def make_name_hash(name='', file_ending='.pt'):
    year, month, day, hour, minute, second = time.localtime()[:6]
    ascii_letters_digits = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    random_hash = ''.join(random.choices(ascii_letters_digits, k=20))
    path = f"{name}_{year}-{month}-{day}_{hour}-{minute}-{second}_{random_hash}{file_ending}"
    return path

class TimeLogger:
    def __init__(self, start=None):
        self.start = self.time() if start is None else start
        self.prev = self.start

    @staticmethod
    def time():
        return time.time()

    def diff(self):
        prev, self.prev = (self.prev, self.time())
        return self.prev - self.start, self.prev - prev

    @staticmethod
    def _hms_from_sec(sec):
        """Hours, minutes, seconds."""
        m, s = divmod(sec, 60)
        h, m = divmod(m, 60)
        return h, m, s

    @staticmethod
    def _hms_str(h, m, s, shorten=True):
        """Hours, minutes, seconds."""
        hs = f"{int(h)}h:"
        ms = f"{int(m)}m:"
        ss = f"{int(s)}s"
        if shorten:
            if h == 0:
                hs = ''
                if m == 0:
                    ms = ''
        return f"{hs}{ms}{ss}"
        # return f"{int(h)}h:{int(m)}m:{int(s)}s"

    def hms_diff(self, shorten=True):
        diff_start, diff_prev = self.diff()
        hms_start = self._hms_from_sec(diff_start)
        hms_prev = self._hms_from_sec(diff_prev)
        return self._hms_str(*hms_start, shorten), self._hms_str(*hms_prev, shorten)


'''
Small changes to the pytorch dataloader. 
Now works with with batches (slicing rather than loop over indexes), 
and can potentially stop workers from shutting down at each batch.
'''
# import random
import warnings
import copy
import torch
import torch.utils.data
import torchtuples


def identity_collate_fn(x):
    return x


class DataLoaderSlice(torch.utils.data.dataloader.DataLoader):
    __doc__ = ("""A hacky version to speed up pytorch's DataLoader.""" + 
               torch.utils.data.dataloader.DataLoader.__doc__)
    # This is a hack that will hopefully be removed from future implementations.
    # The idea is to let the DataSet read a batch at a time, instead of the torch approach of
    # reading one element in the batch at a time (with a loop). For numpy datasets this is much
    # faster.
    # The hack works by setting `self._auto_collation` to False as thils will cause `_MapDatasetFetcher` 
    # to not iterate over the indices in the batch. However, we need to rewrite `self._index_sampler` so
    # we still use a `self.batch_sampler` instead of `self.sampler` when `self._auto_collation` is False.
    def __iter__(self):
        self._done_init = True  # A flag set to change the behavior of _auto_collation
        # self.collate_fn = lambda x: x  # Identity function, because we don't want it do anything.
        self.collate_fn = identity_collate_fn  # Identity function, because we don't want it do anything.
        return super().__iter__()

    @property
    def _auto_collation(self):
        if hasattr(self, '_done_init'):
            return False  # Return false when called by __iter__
        else:
            return super()._auto_collation
    
    @property
    def _index_sampler(self):
        # The actual sampler used for generating indices for `_DatasetFetcher`
        # (see _utils/fetch.py) to read data at each time. This would be
        # `.batch_sampler` if in auto-collation mode, and `.sampler` otherwise.
        # We can't change `.sampler` and `.batch_sampler` attributes for BC
        # reasons.
        if super()._auto_collation: # This will return True even if we have `self._done_init`.
            return self.batch_sampler
        else:
            return self.sampler


if (torch.__version__ >= '1.1.0') and (torch.__version__ < '1.2.0'):
    from torchtuples import _legacy_v1_1_0
    DataLoaderSlice = _legacy_v1_1_0.DataLoaderSlice


class RandomSamplerContinuous(torch.utils.data.dataloader.RandomSampler):
    """Samples elements randomly, without replacement, and continues for ever.

    Arguments:
        data_source (Dataset): dataset to sample from
    """
    def __iter__(self):
        while True:
            for i in iter(torch.randperm(len(self.data_source)).long()):
                yield i


class DatasetTuple(torch.utils.data.Dataset):
    """Dataset where input and target can be tuples.

    Arguments:
        input {tuple or list} -- What is passed to the network (list of x tensors.)
        target {tuple or list} -- Label information passed to the loss function
            (list of y tensors)
    """
    def __init__(self, *data):
        self.data = torchtuples.tuplefy(*data)
        if not self.data.apply(lambda x: type(x) is torch.Tensor).flatten().all():
            warnings.warn("All data is not torch.Tensor. Consider fixing this.")

    def __getitem__(self, index):
        if (not hasattr(index, '__iter__')) and (type(index) is not slice):
            index = [index]
        return self.data.iloc[index]

    def __len__(self):
        lens = self.data.lens().flatten()
        if not lens.all_equal():
            raise RuntimeError("Need all tensors to have same lenght.")
        return lens[0]


class DatasetTupleSingle(DatasetTuple):
    """Like DatasetTuple, but does not read in batches and can therefore be used with
    the regular torch.utils.data.DataLoader.

    Dataset where input and target can be tuples.

    Arguments:
        input {tuple or list} -- What is passed to the network (list of x tensors.)
        target {tuple or list} -- Label information passed to the loss function
            (list of y tensors)
    """
    def __getitem__(self, index):
        return self.data.iloc[index]


class DatasetInputOnly:
    """Class for chaning a Dataset contraining inputs and targets
    to only return the inputs.
    Usefurll for predict methods.
    """
    def __init__(self, dataset):
        self.dataset = dataset

    def __getitem__(self, index):
        return self.dataset[index][0]


def dataloader_input_only(dataloader):
    """Create a new dataloader the only returns the inputs and not the targets.
    Useful when calling predict:

        dataloader = ...
        model.fit(dataloader)
        dl_input = dataloader_input_only(dataloader)
        model.predict(dl_inpt)

    See e.g. MNIST examples code.
    """
    if torch.__version__ <= '1.2.0':
        return _dataloader_input_only_v_less_than_1_2_0(dataloader)

    if type(dataloader.sampler) is not torch.utils.data.sampler.SequentialSampler:
        warnings.warn("Dataloader might not be deterministic!")
    dl = dataloader
    new_dataset = DatasetInputOnly(dl.dataset)
    dl_new = type(dl)(new_dataset, batch_sampler=dl.batch_sampler, num_workers=dl.num_workers,
                      collate_fn=dl.collate_fn, pin_memory=dl.pin_memory, drop_last=dl.drop_last,
                      timeout=dl.timeout, worker_init_fn=dl.worker_init_fn,
                      multiprocessing_context=dl.multiprocessing_context)
    return dl_new

def _dataloader_input_only_v_less_than_1_2_0(dataloader):
    if type(dataloader.sampler) is not torch.utils.data.sampler.SequentialSampler:
        warnings.warn("Dataloader might not be deterministic!")
    dl_new = copy.copy(dataloader)
    dl_new.dataset = DatasetInputOnly(dl_new.dataset)
    return dl_new



# -*- coding: utf-8 -*-

"""Top-level package for torchtuples."""

__author__ = """Haavard Kvamme"""
__email__ = 'haavard.kvamme@gmail.com'
__version__ = '0.1.2'

try:
    import torch
    if torch.__version__ < '1.1.0':
        raise ImportError("""Need a torch version that is at least '1.1.0'""")
except ModuleNotFoundError:
    raise ModuleNotFoundError("""You need to install pytorch! See https://pytorch.org/get-started/locally/""")


import torchtuples.tupletree
from torchtuples.tupletree import TupleTree, tuplefy, to_device, numpy_to_tensor, tensor_to_numpy, make_dataloader
import torchtuples.data
import torchtuples.base
import torchtuples.callbacks
import torchtuples.practical
import torchtuples.optim

from torchtuples.base import Model



import torch
from torch import nn
from torchtuples import tuplefy

def init_embedding(emb):
    """Weight initialization of embeddings (in place).
    Best practise from fastai
    
    Arguments:
        emb {torch.nn.Embedding} -- Embedding
    """
    w = emb.weight.data
    sc = 2 / (w.shape[1]+1)
    w.uniform_(-sc, sc)

def _accuracy(input, target):
    """Accuracy, i.e. mean(input == target)"""
    return input.eq(target.view_as(input)).float().mean()

def accuracy_binary(input, target):
    """Accuracy for binary models on input for logit models in (-inf, inf).
    Do not used for models with sigmoid output activation.
    """
    if len(input.shape) == 1:
        raise NotImplementedError(f"`accuracy_argmax` not implemented for shape {input.shape}")
    assert (target.min() == 0) and (target.max() == 1), 'We have binary classfication so we need 0/1'
    pred = torch.zeros_like(input).to(target.dtype)
    pred[input >= 0.] = 1
    return _accuracy(pred, target)

def accuracy_argmax(input, target):
    """Accuracy after argmax on input for logit models in (-inf, inf).
    Do not used for models with sigmoid/softmax output activation.

    Tupycally used as a metric passed to Model.fit()
    If input is one dimensional, we assume we have binary classification.
    """
    if len(input.shape) == 1:
        raise NotImplementedError(f"`accuracy_argmax` not implemented for shape {input.shape}")
    if input.shape[1] == 1:
        raise NotImplementedError("`accuracy_argmax` not for binary data. See `accuracy_binary` instead.")
    else:
        pred = input.argmax(dim=1, keepdim=True)
    return _accuracy(pred, target)


class DenseVanillaBlock(nn.Module):
    def __init__(self, in_features, out_features, bias=True, batch_norm=True, dropout=0., activation=nn.ReLU,
                 w_init_=lambda w: nn.init.kaiming_normal_(w, nonlinearity='relu')):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features, bias)
        if w_init_:
            w_init_(self.linear.weight.data)
        self.activation = activation()
        self.batch_norm = nn.BatchNorm1d(out_features) if batch_norm else None
        self.dropout = nn.Dropout(dropout) if dropout else None

    def forward(self, input):
        input = self.activation(self.linear(input))
        if self.batch_norm:
            input = self.batch_norm(input)
        if self.dropout:
            input = self.dropout(input)
        return input


class MLPVanilla(nn.Module):
    def __init__(self, in_features, num_nodes, out_features, batch_norm=True, dropout=None, activation=nn.ReLU,
                 output_activation=None, output_bias=True,
                 w_init_=lambda w: nn.init.kaiming_normal_(w, nonlinearity='relu')):
        super().__init__()
        num_nodes = tuplefy(in_features, num_nodes).flatten()
        if not hasattr(dropout, '__iter__'):
            dropout = [dropout for _ in range(len(num_nodes)-1)]
        net = []
        for n_in, n_out, p in zip(num_nodes[:-1], num_nodes[1:], dropout):
            net.append(DenseVanillaBlock(n_in, n_out, True, batch_norm, p, activation, w_init_))
        net.append(nn.Linear(num_nodes[-1], out_features, output_bias))
        if output_activation:
            net.append(output_activation)
        self.net = nn.Sequential(*net)

    def forward(self, input):
        return self.net(input)


class EntityEmbeddings(nn.Module):
    def __init__(self, num_embeddings, embedding_dims, dropout=0.):
        super().__init__()
        if not hasattr(num_embeddings, '__iter__'):
            num_embeddings = [num_embeddings]
        if not hasattr(embedding_dims, '__iter__'):
            embedding_dims = [embedding_dims]
        if len(num_embeddings) != len(embedding_dims):
            raise ValueError("Need 'num_embeddings' and 'embedding_dims' to have the same length")
        self.embeddings = nn.ModuleList()
        for n_emb, emb_dim in zip(num_embeddings, embedding_dims):
            emb = nn.Embedding(n_emb, emb_dim)
            init_embedding(emb)
            self.embeddings.append(emb)
        self.dropout = nn.Dropout(dropout) if dropout else None
    
    def forward(self, input):
        if input.shape[1] != len(self.embeddings):
            raise RuntimeError(f"Got input of shape '{input.shape}', but need dim 1 to be {len(self.embeddings)}.")
        input = [emb(input[:, i]) for i, emb in enumerate(self.embeddings)]
        input = torch.cat(input, 1)
        if self.dropout:
            input = self.dropout(input)
        return input


class MixedInputMLP(nn.Module):
    def __init__(self, in_features, num_embeddings, embedding_dims, num_nodes, out_features,
                 batch_norm=True, dropout=None, activation=nn.ReLU, dropout_embedding=0.,
                 output_activation=None, output_bias=True,
                 w_init_=lambda w: nn.init.kaiming_normal_(w, nonlinearity='relu')):
        super().__init__()
        self.embeddings = EntityEmbeddings(num_embeddings, embedding_dims, dropout_embedding)
        input_mlp = in_features + sum(embedding_dims)
        self.mlp = MLPVanilla(input_mlp, num_nodes, out_features, batch_norm, dropout, activation,
                              output_activation, output_bias, w_init_)

    def forward(self, input_numeric, input_categoric):
        input = torch.cat([input_numeric, self.embeddings(input_categoric)], 1)
        return self.mlp(input)



"""Contains some implenentations needed for torch.__version__ == 1.1.0.

Code copied from pytorch source. Have just changed a couple of lines.

We have made _DataLoaderIterSlice which use a different '_worker_loop' and 'collate_fn'.
"""
import torch
import torch.multiprocessing as multiprocessing
from torch.utils.data import _utils
import threading
from torch._six import queue
from torch.utils.data._utils.worker import *


class DataLoaderSlice(torch.utils.data.dataloader.DataLoader):
    r"""
    Like DataLoader but works on batches instead of iterating
    through the batch.

    Data loader. Combines a dataset and a sampler, and provides
    single- or multi-process iterators over the dataset.

    Arguments:
        dataset (Dataset): dataset from which to load the data.
        batch_size (int, optional): how many samples per batch to load
            (default: 1).
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: False).
        sampler (Sampler, optional): defines the strategy to draw samples from
            the dataset. If specified, ``shuffle`` must be False.
        batch_sampler (Sampler, optional): like sampler, but returns a batch of
            indices at a time. Mutually exclusive with batch_size, shuffle,
            sampler, and drop_last.
        num_workers (int, optional): how many subprocesses to use for data
            loading. 0 means that the data will be loaded in the main process.
            (default: 0)
        collate_fn (callable, optional): merges a list of samples to form a mini-batch.
            If None, we simply pass the the in put to collate_fn.
        pin_memory (bool, optional): If ``True``, the data loader will copy tensors
            into CUDA pinned memory before returning them.
        drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
            if the dataset size is not divisible by the batch size. If ``False`` and
            the size of dataset is not divisible by the batch size, then the last batch
            will be smaller. (default: False)
        timeout (numeric, optional): if positive, the timeout value for collecting a batch
            from workers. Should always be non-negative. (default: 0)
        worker_init_fn (callable, optional): If not None, this will be called on each
            worker subprocess with the worker id (an int in ``[0, num_workers - 1]``) as
            input, after seeding and before data loading. (default: None)

    .. note:: By default, each worker will have its PyTorch seed set to
              ``base_seed + worker_id``, where ``base_seed`` is a long generated
              by main process using its RNG. However, seeds for other libraies
              may be duplicated upon initializing workers (w.g., NumPy), causing
              each worker to return identical random numbers. (See
              :ref:`dataloader-workers-random-seed` section in FAQ.) You may
              use ``torch.initial_seed()`` to access the PyTorch seed for each
              worker in :attr:`worker_init_fn`, and use it to set other seeds
              before data loading.

    .. warning:: If ``spawn`` start method is used, :attr:`worker_init_fn` cannot be an
                 unpicklable object, e.g., a lambda function.
    """
    def __init__(self, dataset, batch_size=1, shuffle=False, sampler=None, batch_sampler=None,
                 num_workers=0, collate_fn=None, pin_memory=False,
                 drop_last=False, timeout=0, worker_init_fn=None):
        if collate_fn is None:
            collate_fn = DataLoaderSlice._identity
        super().__init__(dataset, batch_size, shuffle, sampler, batch_sampler, num_workers,
                         collate_fn, pin_memory, drop_last, timeout, worker_init_fn)

    def __iter__(self):
        return _DataLoaderIterSlice(self)

    @staticmethod
    def _identity(x):
        '''Function returning x'''
        return x




##################################################################################################
##################################################################################################
## The following is copyed and slightly changed from pytorch source.
##################################################################################################

# This function used to be defined in this file. However, it was moved to
# _utils/collate.py. Although it is rather hard to access this from user land
# (one has to explicitly directly `import torch.utils.data.dataloader`), there
# probably is user code out there using it. This aliasing maintains BC in this
# aspect.
# default_collate = _utils.collate.default_collate

def _worker_loop(dataset, index_queue, data_queue, done_event, collate_fn, seed, init_fn, worker_id):
    # See NOTE [ Data Loader Multiprocessing Shutdown Logic ] for details on the
    # logic of this function.

    try:
        collate._use_shared_memory = True

        # Intialize C side signal handlers for SIGBUS and SIGSEGV. Python signal
        # module's handlers are executed after Python returns from C low-level
        # handlers, likely when the same fatal signal had already happened
        # again.
        # https://docs.python.org/3/library/signal.html#execution-of-python-signal-handlers
        signal_handling._set_worker_signal_handlers()

        torch.set_num_threads(1)
        random.seed(seed)
        torch.manual_seed(seed)

        data_queue.cancel_join_thread()

        init_exception = None

        if init_fn is not None:
            try:
                init_fn(worker_id)
            except Exception:
                init_exception = ExceptionWrapper(sys.exc_info())

        watchdog = ManagerWatchdog()

        while watchdog.is_alive():
            try:
                r = index_queue.get(timeout=MP_STATUS_CHECK_INTERVAL)
            except queue.Empty:
                continue
            if r is None:
                # Received the final signal
                assert done_event.is_set()
                return
            elif done_event.is_set():
                # Done event is set. But I haven't received the final signal
                # (None) yet. I will keep continuing until get it, and skip the
                # processing steps.
                continue
            idx, batch_indices = r
            try:
                if init_exception is not None:
                    samples = init_exception
                    init_exception = None
                else:
                    # samples = collate_fn([dataset[i] for i in batch_indices])
                    samples = collate_fn(dataset[batch_indices]) # changed in torchtuples
            except Exception:
                # It is important that we don't store exc_info in a variable,
                # see NOTE [ Python Traceback Reference Cycle Problem ]
                data_queue.put((idx, ExceptionWrapper(sys.exc_info())))
            else:
                data_queue.put((idx, samples))
                del samples
    except KeyboardInterrupt:
        # Main process will raise KeyboardInterrupt anyways.
        pass


class _DataLoaderIterSlice(object):
    r"""Iterates once over the DataLoader's dataset, as specified by the sampler"""

    # NOTE [ Data Loader Multiprocessing Shutdown Logic ]
    #
    # Preliminary:
    #
    # Our data model looks like this (queues are indicated with curly brackets):
    #
    #                main process                              ||
    #                     |                                    ||
    #               {index_queue}                              ||
    #                     |                                    ||
    #              worker processes                            ||     DATA
    #                     |                                    ||
    #            {worker_result_queue}                         ||     FLOW
    #                     |                                    ||
    #      pin_memory_thread of main process                   ||   DIRECTION
    #                     |                                    ||
    #               {data_queue}                               ||
    #                     |                                    ||
    #                data output                               \/
    #
    # P.S. `worker_result_queue` and `pin_memory_thread` part may be omitted if
    #      `pin_memory=False`.
    #
    #
    # Terminating multiprocessing logic requires very careful design. In
    # particular, we need to make sure that
    #
    #   1. The iterator gracefully exits the workers when its last reference is
    #      gone or it is depleted.
    #
    #      In this case, the workers should be gracefully exited because the
    #      main process may still need to continue to run, and we want cleaning
    #      up code in the workers to be executed (e.g., releasing GPU memory).
    #      Naturally, we implement the shutdown logic in `__del__` of
    #      DataLoaderIterator.
    #
    #      We delay the discussion on the logic in this case until later.
    #
    #   2. The iterator exits the workers when the loader process and/or worker
    #      processes exits normally or with error.
    #
    #      We set all workers and `pin_memory_thread` to have `daemon=True`.
    #
    #      You may ask, why can't we make the workers non-daemonic, and
    #      gracefully exit using the same logic as we have in `__del__` when the
    #      iterator gets deleted (see 1 above)?
    #
    #      First of all, `__del__` is **not** guaranteed to be called when
    #      interpreter exits. Even if it is called, by the time it executes,
    #      many Python core library resources may alreay be freed, and even
    #      simple things like acquiring an internal lock of a queue may hang.
    #      Therefore, in this case, we actually need to prevent `__del__` from
    #      being executed, and rely on the automatic termination of daemonic
    #      children. Thus, we register an `atexit` hook that sets a global flag
    #      `_utils.python_exit_status`. Since `atexit` hooks are executed in the
    #      reverse order of registration, we are guaranteed that this flag is
    #      set before library resources we use are freed. (Hooks freeing those
    #      resources are registered at importing the Python core libraries at
    #      the top of this file.) So in `__del__`, we check if
    #      `_utils.python_exit_status` is set or `None` (freed), and perform
    #      no-op if so.
    #
    #      Another problem with `__del__` is also related to the library cleanup
    #      calls. When a process ends, it shuts the all its daemonic children
    #      down with a SIGTERM (instead of joining them without a timeout).
    #      Simiarly for threads, but by a different mechanism. This fact,
    #      together with a few implementation details of multiprocessing, forces
    #      us to make workers daemonic. All of our problems arise when a
    #      DataLoader is used in a subprocess, and are caused by multiprocessing
    #      code which looks more or less like this:
    #
    #          try:
    #              your_function_using_a_dataloader()
    #          finally:
    #              multiprocessing.util._exit_function()
    #
    #      The joining/termination mentioned above happens inside
    #      `_exit_function()`. Now, if `your_function_using_a_dataloader()`
    #      throws, the stack trace stored in the exception will prevent the
    #      frame which uses `DataLoaderIter` to be freed. If the frame has any
    #      reference to the `DataLoaderIter` (e.g., in a method of the iter),
    #      its  `__del__`, which starts the shutdown procedure, will not be
    #      called. That, in turn, means that workers aren't notified. Attempting
    #      to join in `_exit_function` will then result in a hang.
    #
    #      For context, `_exit_function` is also registered as an `atexit` call.
    #      So it is unclear to me (@ssnl) why this is needed in a finally block.
    #      The code dates back to 2008 and there is no comment on the original
    #      PEP 371 or patch https://bugs.python.org/issue3050 (containing both
    #      the finally block and the `atexit` registration) that explains this.
    #
    #      Another choice is to just shutdown workers with logic in 1 above
    #      whenever we see an error in `next`. This isn't ideal because
    #        a. It prevents users from using try-catch to resume data loading.
    #        b. It doesn't prevent hanging if users have references to the
    #           iterator.
    #
    #   3. All processes exit if any of them die unexpectedly by fatal signals.
    #
    #      As shown above, the workers are set as daemonic children of the main
    #      process. However, automatic cleaning-up of such child processes only
    #      happens if the parent process exits gracefully (e.g., not via fatal
    #      signals like SIGKILL). So we must ensure that each process will exit
    #      even the process that should send/receive data to/from it were
    #      killed, i.e.,
    #
    #        a. A process won't hang when getting from a queue.
    #
    #           Even with carefully designed data dependencies (i.e., a `put()`
    #           always corresponding to a `get()`), hanging on `get()` can still
    #           happen when data in queue is corrupted (e.g., due to
    #           `cancel_join_thread` or unexpected exit).
    #
    #           For child exit, we set a timeout whenever we try to get data
    #           from `data_queue`, and check the workers' status on each timeout
    #           and error.
    #           See `_DataLoaderiter._get_batch()` and
    #           `_DataLoaderiter._try_get_batch()` for details.
    #
    #           Additionally, for child exit on non-Windows platforms, we also
    #           register a SIGCHLD handler (which is supported on Windows) on
    #           the main process, which checks if any of the workers fail in the
    #           (Python) handler. This is more efficient and faster in detecting
    #           worker failures, compared to only using the above mechanism.
    #           See `DataLoader.cpp` and `_utils/signal_handling.py` for details.
    #
    #           For `.get()` calls where the sender(s) is not the workers, we
    #           guard them with timeouts, and check the status of the sender
    #           when timeout happens:
    #             + in the workers, the `_utils.worker.ManagerWatchdog` class
    #               checks the status of the main process.
    #             + if `pin_memory=True`, when getting from `pin_memory_thread`,
    #               check `pin_memory_thread` status periodically until `.get()`
    #               returns or see that `pin_memory_thread` died.
    #
    #        b. A process won't hang when putting into a queue;
    #
    #           We use `mp.Queue` which has a separate background thread to put
    #           objects from an unbounded buffer array. The background thread is
    #           daemonic and usually automatically joined when the process
    #           exits.
    #
    #           However, in case that the receiver has ended abruptly while
    #           reading from the pipe, the join will hang forever. Therefore,
    #           for both `worker_result_queue` (worker -> main process/pin_memory_thread)
    #           and each `index_queue` (main process -> worker), we use
    #           `q.cancel_join_thread()` in sender process before any `q.put` to
    #           prevent this automatic join.
    #
    #           Moreover, having all queues called `cancel_join_thread` makes
    #           implementing graceful shutdown logic in `__del__` much easier.
    #           It won't need to get from any queue, which would also need to be
    #           guarded by periodic status checks.
    #
    #           Note that this may leave corrupted data in the queue, but we
    #           don't care about the data anyways once we are shutting down.
    #
    #
    # Now let's get back to 1:
    #   how we gracefully exit the workers when the last reference to the
    #   iterator is gone.
    #
    # To achieve this, we implement the following logic along with the design
    # choices mentioned above:
    #
    # [worker processes]
    #   While loader process is alive:
    #     Get from index_queue.
    #       If got a `None`, exit.
    #       If get anything else,
    #          Check `done_event`.
    #            If set, continue to next iteration
    #                    i.e., keep getting until see the `None`, then exit.
    #            Otherwise, process data.
    #       If timed out,
    #          No matter `done_event` is set (still need to see `None`) or not,
    #          must continue to next iteration .
    #
    # [pin_memory_thread]
    #   # No need to check main thread. If this thread is alive, the main loader
    #   # thread must be alive, because this thread is set as daemonic.
    #   While True:
    #     Get from index_queue.
    #       If got a `None`, exit.
    #       If get anything else,
    #          Check `done_event`.
    #            If set, continue to next iteration
    #                    i.e., keep getting until see the `None`, then exit.
    #            Otherwise, process data.
    #
    #   NOTE: we don't check the status of the main thread because
    #           1. if the process is killed by fatal signal, `pin_memory_thread`
    #              ends.
    #           2. in other cases, either the cleaning-up in __del__ or the
    #              automatic exit of daemonic thread will take care of it.
    #              This won't busy-wait either because `.get(timeout)` does not
    #              busy-wait.
    #
    # [main process]
    #   In the DataLoader Iter's `__del__`
    #     a. Set `done_event` (shared with `pin_memory_thread` and workers).
    #
    #        Note: from here on, the workers & `pin_memory_thread` may exit at
    #              any time after they receive `None`.
    #
    #     b. Exit `pin_memory_thread`
    #          i.   Put `None` in `worker_result_queue`.
    #          ii.  Join the `pin_memory_thread`.
    #
    #     c. Exit the workers.
    #          i.   Put `None` in each worker's `index_queue`.
    #          ii.  Join the workers.
    #
    #        NOTE: This has to be after (b) because it may leave corrupted data
    #              in `worker_result_queue`, which `pin_memory_thread` reads
    #              from.
    #
    #   NOTE: If `pin_memory=False`, there is no `pin_memory_thread` and (b)
    #         can be omitted
    #
    # NB: `done_event`s isn't strictly needed. E.g., we can just check for
    #     `None` from `index_queue`, but it allows us to skip wasting resources
    #     processing indices already in `index_queue` if we are already shutting
    #     down.

    def __init__(self, loader):
        self.dataset = loader.dataset
        self.collate_fn = loader.collate_fn
        self.batch_sampler = loader.batch_sampler
        self.num_workers = loader.num_workers
        self.pin_memory = loader.pin_memory and torch.cuda.is_available()
        self.timeout = loader.timeout

        self.sample_iter = iter(self.batch_sampler)

        base_seed = torch.LongTensor(1).random_().item()

        if self.num_workers > 0:
            self.worker_init_fn = loader.worker_init_fn
            self.worker_queue_idx = 0
            self.worker_result_queue = multiprocessing.Queue()
            self.batches_outstanding = 0
            self.worker_pids_set = False
            self.shutdown = False
            self.send_idx = 0
            self.rcvd_idx = 0
            self.reorder_dict = {}
            self.done_event = multiprocessing.Event()

            self.index_queues = []
            self.workers = []
            for i in range(self.num_workers):
                index_queue = multiprocessing.Queue()
                index_queue.cancel_join_thread()
                w = multiprocessing.Process(
                    # target=_utils.worker._worker_loop,
                    target=_worker_loop,
                    args=(self.dataset, index_queue,
                          self.worker_result_queue, self.done_event,
                          self.collate_fn, base_seed + i,
                          self.worker_init_fn, i))
                w.daemon = True
                # NB: Process.start() actually take some time as it needs to
                #     start a process and pass the arguments over via a pipe.
                #     Therefore, we only add a worker to self.workers list after
                #     it started, so that we do not call .join() if program dies
                #     before it starts, and __del__ tries to join but will get:
                #     AssertionError: can only join a started process.
                w.start()
                self.index_queues.append(index_queue)
                self.workers.append(w)

            if self.pin_memory:
                self.data_queue = queue.Queue()
                pin_memory_thread = threading.Thread(
                    target=_utils.pin_memory._pin_memory_loop,
                    args=(self.worker_result_queue, self.data_queue,
                          torch.cuda.current_device(), self.done_event))
                pin_memory_thread.daemon = True
                pin_memory_thread.start()
                # Similar to workers (see comment above), we only register
                # pin_memory_thread once it is started.
                self.pin_memory_thread = pin_memory_thread
            else:
                self.data_queue = self.worker_result_queue

            _utils.signal_handling._set_worker_pids(id(self), tuple(w.pid for w in self.workers))
            _utils.signal_handling._set_SIGCHLD_handler()
            self.worker_pids_set = True

            # prime the prefetch loop
            for _ in range(2 * self.num_workers):
                self._put_indices()

    def __len__(self):
        return len(self.batch_sampler)

    def _try_get_batch(self, timeout=_utils.MP_STATUS_CHECK_INTERVAL):
        # Tries to fetch data from `data_queue` for a given timeout. This can
        # also be used as inner loop of fetching without timeout, with the
        # sender status as the loop condition.
        #
        # This raises a `RuntimeError` if any worker died expectedly. This error
        # can come from either the SIGCHLD handler in `_utils/signal_handling.py`
        # (only for non-Windows platforms), or the manual check below on errors
        # and timeouts.
        #
        # Returns a 2-tuple:
        #   (bool: whether successfully get data, any: data if successful else None)
        try:
            data = self.data_queue.get(timeout=timeout)
            return (True, data)
        except Exception as e:
            # At timeout and error, we manually check whether any worker has
            # failed. Note that this is the only mechanism for Windows to detect
            # worker failures.
            if not all(w.is_alive() for w in self.workers):
                pids_str = ', '.join(str(w.pid) for w in self.workers if not w.is_alive())
                raise RuntimeError('DataLoader worker (pid(s) {}) exited unexpectedly'.format(pids_str))
            if isinstance(e, queue.Empty):
                return (False, None)
            raise

    def _get_batch(self):
        # Fetches data from `self.data_queue`.
        #
        # We check workers' status every `MP_STATUS_CHECK_INTERVAL` seconds,
        # which we achieve by running `self._try_get_batch(timeout=MP_STATUS_CHECK_INTERVAL)`
        # in a loop. This is the only mechanism to detect worker failures for
        # Windows. For other platforms, a SIGCHLD handler is also used for
        # worker failure detection.
        #
        # If `pin_memory=True`, we also need check if `pin_memory_thread` had
        # died at timeouts.
        if self.timeout > 0:
            success, data = self._try_get_batch(self.timeout)
            if success:
                return data
            else:
                raise RuntimeError('DataLoader timed out after {} seconds'.format(self.timeout))
        elif self.pin_memory:
            while self.pin_memory_thread.is_alive():
                success, data = self._try_get_batch()
                if success:
                    return data
            else:
                # while condition is false, i.e., pin_memory_thread died.
                raise RuntimeError('Pin memory thread exited unexpectedly')
            # In this case, `self.data_queue` is a `queue.Queue`,. But we don't
            # need to call `.task_done()` because we don't use `.join()`.
        else:
            while True:
                success, data = self._try_get_batch()
                if success:
                    return data

    def __next__(self):
        if self.num_workers == 0:  # same-process loading
            indices = next(self.sample_iter)  # may raise StopIteration
            # batch = self.collate_fn([self.dataset[i] for i in indices])
            batch = self.collate_fn(self.dataset[indices]) # changed in torchtuples
            if self.pin_memory:
                batch = _utils.pin_memory.pin_memory_batch(batch)
            return batch

        # check if the next sample has already been generated
        if self.rcvd_idx in self.reorder_dict:
            batch = self.reorder_dict.pop(self.rcvd_idx)
            return self._process_next_batch(batch)

        if self.batches_outstanding == 0:
            self._shutdown_workers()
            raise StopIteration

        while True:
            assert (not self.shutdown and self.batches_outstanding > 0)
            idx, batch = self._get_batch()
            self.batches_outstanding -= 1
            if idx != self.rcvd_idx:
                # store out-of-order samples
                self.reorder_dict[idx] = batch
                continue
            return self._process_next_batch(batch)

    next = __next__  # Python 2 compatibility

    def __iter__(self):
        return self

    def _put_indices(self):
        assert self.batches_outstanding < 2 * self.num_workers
        indices = next(self.sample_iter, None)
        if indices is None:
            return
        self.index_queues[self.worker_queue_idx].put((self.send_idx, indices))
        self.worker_queue_idx = (self.worker_queue_idx + 1) % self.num_workers
        self.batches_outstanding += 1
        self.send_idx += 1

    def _process_next_batch(self, batch):
        self.rcvd_idx += 1
        self._put_indices()
        if isinstance(batch, _utils.ExceptionWrapper):
            # make multiline KeyError msg readable by working around
            # a python bug https://bugs.python.org/issue2651
            if batch.exc_type == KeyError and "\n" in batch.exc_msg:
                raise Exception("KeyError:" + batch.exc_msg)
            else:
                raise batch.exc_type(batch.exc_msg)
        return batch

    def __getstate__(self):
        # TODO: add limited pickling support for sharing an iterator
        # across multiple threads for HOGWILD.
        # Probably the best way to do this is by moving the sample pushing
        # to a separate thread and then just sharing the data queue
        # but signalling the end is tricky without a non-blocking API
        raise NotImplementedError("_DataLoaderIter cannot be pickled")

    def _shutdown_workers(self):
        # See NOTE [ Data Loader Multiprocessing Shutdown Logic ] for details on
        # the logic of this function.
        python_exit_status = _utils.python_exit_status
        if python_exit_status is True or python_exit_status is None:
            # See (2) of the note. If Python is shutting down, do no-op.
            return
        # Normal exit when last reference is gone / iterator is depleted.
        # See (1) and the second half of the note.
        if not self.shutdown:
            self.shutdown = True
            try:
                self.done_event.set()

                # Exit `pin_memory_thread` first because exiting workers may leave
                # corrupted data in `worker_result_queue` which `pin_memory_thread`
                # reads from.
                if hasattr(self, 'pin_memory_thread'):
                    # Use hasattr in case error happens before we set the attribute.
                    # First time do `worker_result_queue.put` in this process.

                    # `cancel_join_thread` in case that `pin_memory_thread` exited.
                    self.worker_result_queue.cancel_join_thread()
                    self.worker_result_queue.put(None)
                    self.pin_memory_thread.join()
                    # Indicate that no more data will be put on this queue by the
                    # current process. This **must** be called after
                    # `pin_memory_thread` is joined because that thread shares the
                    # same pipe handles with this loader thread. If the handle is
                    # closed, Py3 will error in this case, but Py2 will just time
                    # out even if there is data in the queue.
                    self.worker_result_queue.close()

                # Exit workers now.
                for q in self.index_queues:
                    q.put(None)
                    # Indicate that no more data will be put on this queue by the
                    # current process.
                    q.close()
                for w in self.workers:
                    w.join()
            finally:
                # Even though all this function does is putting into queues that
                # we have called `cancel_join_thread` on, weird things can
                # happen when a worker is killed by a signal, e.g., hanging in
                # `Event.set()`. So we need to guard this with SIGCHLD handler,
                # and remove pids from the C side data structure only at the
                # end.
                #
                # FIXME: Unfortunately, for Windows, we are missing a worker
                #        error detection mechanism here in this function, as it
                #        doesn't provide a SIGCHLD handler.
                if self.worker_pids_set:
                    _utils.signal_handling._remove_worker_pids(id(self))
                    self.worker_pids_set = False

    def __del__(self):
        if self.num_workers > 0:
            self._shutdown_workers()



from torch import optim
import torchtuples.callbacks as cb

class OptimWrap(cb.CallbackHandler):
    """Wraps a torch.optim.Optimizer object so we can call some extra methods on it.
    The torch.optim.Optimizer can be obtained through the property 'optimizer'. 
    """
    def __init__(self, optimizer, callbacks=None):
        self.optimizer = optimizer
        if callbacks is None:
            callbacks = {}
        super().__init__(callbacks)
    
    def step(self, closure=None):
        return self.optimizer.step(closure)

    def zero_grad(self):
        return self.optimizer.zero_grad()

    @property
    def param_groups(self):
        return self.optimizer.param_groups

    def add_param_group(self, param_group):
        return self.optimizer.add_param_group(param_group)

    def load_state_dict(self, state_dict):
        return self.optimizer.load_state_dict(state_dict)

    def state_dict(self):
        return self.optimizer.state_dict()
    
    def set(self, key, val):
        for param_group in self.optimizer.param_groups:
            param_group[key] = val
        # return self

    # def set_decoupled_weight_decay(self, wd):
    #     """Set decoupled weight decay"""
    #     return self.set('decoupled_weight_decay', wd)

    def set_lr(self, lr):
        """Sets 'initial_lr' and 'lr' to value 'lr'"""
        self.set('initial_lr', lr)
        self.set('lr', lr)
        # return self

    def set_momentum(self, momentum):
        first_gr = self.optimizer.parameter_groups[0]
        if 'betas' in first_gr:
            for param_group in self.optimizer.param_groups:
                param_group['betas'] = (momentum, param_group['betas'][1])
        elif 'momentum' in first_gr:
            self.set('momentum', momentum)
        else:
            raise ValueError("No momentum found")
        # return self

    def set_beta(self, beta):
        first_gr = self.optimizer.parameter_groups[0]
        if 'betas' in first_gr:
            for param_group in self.optimizer.param_groups:
                param_group['betas'] = (param_group['betas'][0], beta)
        elif 'alpha' in first_gr:
            self.set('alpha', beta)
        else:
            raise ValueError("No beta found")
        # return self
    
    def drop_scheduler(self):
        pass

    def reinitialize(self, params=None, **kwargs):
        print('Optimizer is not reinitialized. Need to do this manually')
        return NotImplemented

class OptimWrapReinit(OptimWrap):
    optim_func = NotImplemented
    init_args = NotImplemented

    def __init__(self, callbacks=None, params=None):
        super().__init__(None, callbacks)
        if params is not None:
            self.init_optimizer(params)

    def __call__(self, params):
        self.init_optimizer(params)
        return self

    def init_optimizer(self, params):
        call_args = set(self.optim_func.__init__.__code__.co_varnames[1:])
        init_args = {name: self.init_args[name] for name in (call_args & self.init_args.keys())}
        self.optimizer = self.optim_func(params, **init_args)

    def reinitialize(self, params=None, **kwargs):
        if params is None:
            if hasattr(self, 'model'):
                params = self.model.net.parameters()
        init_args = self.init_args.copy()
        init_args.update(kwargs)
        return self._constructor(**init_args, params=params)

    @property
    def _constructor(self):
        raise NotImplementedError


class SGD(OptimWrapReinit):
    r"""Wrapper to torch.optim.SGD where 'params' are not needed.

    Implements stochastic gradient descent (optionally with momentum).

    Nesterov momentum is based on the formula from
    `On the importance of initialization and momentum in deep learning`__.

    Args:
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups. If not specified, it will get parameters from model.

    __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf

    .. note::
        The implementation of SGD with Momentum/Nesterov subtly differs from
        Sutskever et. al. and implementations in some other frameworks.

        Considering the specific case of Momentum, the update can be written as

        .. math::
                  v = \rho * v + g \\
                  p = p - lr * v

        where p, g, v and :math:`\rho` denote the parameters, gradient,
        velocity, and momentum respectively.

        This is in contrast to Sutskever et. al. and
        other frameworks which employ an update of the form

        .. math::
             v = \rho * v + lr * g \\
             p = p - v

        The Nesterov version is analogously modified.
    """
    optim_func = optim.SGD
    def __init__(self, lr=1e-2, momentum=0, dampening=0,
                 weight_decay=0, nesterov=False, params=None):
        self.init_args = dict(lr=lr, momentum=momentum, dampening=dampening, weight_decay=weight_decay,
                              nesterov=nesterov)
        super().__init__(None, params)

    @property
    def _constructor(self):
        return SGD


class RMSprop(OptimWrapReinit):
    """Wrapper to torch.optim.RMSprop where 'params' are not 'needed'.

    Implements RMSprop algorithm.

    Proposed by G. Hinton in his
    `course <http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.

    The centered version first appears in `Generating Sequences
    With Recurrent Neural Networks <https://arxiv.org/pdf/1308.0850v5.pdf>`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-2)
        momentum (float, optional): momentum factor (default: 0)
        alpha (float, optional): smoothing constant (default: 0.99)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        centered (bool, optional) : if ``True``, compute the centered RMSProp,
            the gradient is normalized by an estimation of its variance
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

    """
    optim_func = optim.RMSprop
    def __init__(self, lr=1e-2, alpha=0.99, eps=1e-8, weight_decay=0, momentum=0, centered=False,
                 params=None):
        self.init_args = dict(lr=lr, alpha=alpha, eps=eps, weight_decay=weight_decay, momentum=momentum,
                              centered=centered)
        super().__init__(None, params)

    @property
    def _constructor(self):
        return RMSprop

class Adam(OptimWrapReinit):
    r"""Wrapper to torch.optim.Adam where 'params' are not 'needed'.

    Implements Adam algorithm.

    It has been proposed in `Adam: A Method for Stochastic Optimization`_.

    Arguments:
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups. If not specified, it will get parameters from model.

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """
    optim_func = optim.Adam
    def __init__(self, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, amsgrad=False,
                 params=None):
        self.init_args = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad)
        super().__init__(None, params)

    @property
    def _constructor(self):
        return Adam


class OptimizerDecoupledWeightDecay(OptimWrapReinit):
    def __init__(self, decoupled_weight_decay=0, dwd_normalize=False,
                 nb_epochs=None, params=None, callbacks=None):
        callbacks = callbacks if callbacks else {}
        if 'decoupled_weight_decay' in callbacks.keys():
            raise ValueError("decoupled_weight_decay allreday exists")
        self.decoupled_weight_decay = cb.DecoupledWeightDecay(decoupled_weight_decay, dwd_normalize,
                                                              nb_epochs)
        callbacks['decoupled_weight_decay'] = self.decoupled_weight_decay
        super().__init__(callbacks, params)

    def init_optimizer(self, params):
        return super().init_optimizer(params)
        # self.set_decoupled_weight_decay(self.init_args['decoupled_weight_decay'])

    def step(self, closure=None):
        if not hasattr(self, 'model'):
            raise RuntimeError("Optimizer with decoupled weight decay needs assignent of a 'Model' to function properly")
        return super().step(closure)


class AdamW(OptimizerDecoupledWeightDecay):
    optim_func = optim.Adam
    def __init__(self, lr=1e-3, betas=(0.9, 0.99), decoupled_weight_decay=0., dwd_normalize=False,
                 nb_epochs=None, eps=1e-8, params=None):
        self.init_args = dict(lr=lr, betas=betas, eps=eps, decoupled_weight_decay=decoupled_weight_decay,
                              dwd_normalize=dwd_normalize, nb_epochs=nb_epochs)
        super().__init__(decoupled_weight_decay, dwd_normalize, nb_epochs, params)

    @property
    def _constructor(self):
        return AdamW


class AdamWR(OptimizerDecoupledWeightDecay):
    """Adam with decoupled weight decay and warm restarts
    Eta is multiplied with this learning rate.
    
    Keyword Arguments:
        lr {float} -- Learning rate. (default: {1e-3})
        betas {tuple} -- Betas in Adam. (default: {(0.9, 0.99)})
        wd {float} -- Decoupled weight decay (default: {0.})
        wd_normalize {bool} -- Normalized weight decay. (default: {True})
        cycle_len {int} -- Number of epochs in each cycle. (default: {1})
        cycle_multiplier {int} -- After each cycle multiply cycle len with this (default: {2})
        cycle_eta_multiplier {[type]} -- After each cycle multiply eta_max with this (default: {1.})
        eta_min {int} -- Min eta  (default: {0})
        params {[type]} -- torch net parameters (default: {None})
        eps {[type]} --  (default: {1e-8})
    """
    optim_func = optim.Adam
    def __init__(self, lr=1e-3, betas=(0.9, 0.99), decoupled_weight_decay=0., dwd_normalize=True, 
                 cycle_len=1, cycle_multiplier=2, cycle_eta_multiplier=1.,
                 eta_min=0, params=None, eps=1e-8):

        self.init_args = dict(lr=lr, betas=betas, eps=eps, decoupled_weight_decay=decoupled_weight_decay,
                              dwd_normalize=dwd_normalize, cycle_len=cycle_len, cycle_multiplier=cycle_multiplier,
                              cycle_eta_multiplier=cycle_eta_multiplier, eta_min=eta_min)
        self.lr_scheduler = cb.LRCosineAnnealing(cycle_len, cycle_multiplier,
                                                 cycle_eta_multiplier,eta_min)
        callbacks = {'lr_scheduler': self.lr_scheduler}
        nb_epochs = self.lr_scheduler.get_cycle_len if dwd_normalize else None
        super().__init__(decoupled_weight_decay, dwd_normalize, nb_epochs, params, callbacks)
    
    def drop_scheduler(self):
        self.callbacks.pop('lr_scheduler')
        self.decoupled_weight_decay.nb_epochs = 1

    @property
    def _constructor(self):
        return AdamWR




import math
import numpy as np
from torch.optim import Optimizer
try:
    import pandas as pd
except:
    pass


class LRSchedulerBatch(object):
    def __init__(self, optimizer, batch_iter=-1):
        # if not isinstance(optimizer, Optimizer):
        #     raise TypeError(f'{type(optimizer).__name__} is not an Optimizer')
        self.optimizer = optimizer
        if batch_iter == -1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError(f"""param 'initial_lr' is not specified
                                   in param_groups[{i}] when resuming an optimizer""")
        self.batch_iter = batch_iter + (batch_iter == -1)

    def state_dict(self):
        """Returns the state of the scheduler as a :class:`dict`.

        It contains an entry for every variable in self.__dict__ which
        is not the optimizer.
        """
        return {key: value for key, value in self.__dict__.items() if key != 'optimizer'}

    def load_state_dict(self, state_dict):
        """Loads the schedulers state.

        Arguments:
            state_dict (dict): scheduler state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        self.__dict__.update(state_dict)
    
    def get_lr(self):
        raise NotImplementedError

    def step(self, score=None, batch_iter=None):
        if batch_iter is not None:
            self.batch_iter = batch_iter
        for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = lr
        self.batch_iter += 1


class LRFinderScheduler(LRSchedulerBatch):
    def __init__(self, optimizer, lr_lower=1e-7, lr_upper=10., n_steps=100):
        super().__init__(optimizer, batch_iter=-1)
        ratio = np.power(lr_upper/lr_lower, 1/(n_steps-1))
        self.lrs = lr_lower * np.power(ratio, np.arange(n_steps))
    
    def get_lr(self):
        return [self.lrs[self.batch_iter]] * len(self.optimizer.param_groups)
        

class LRBatchCosineAnnealing(LRSchedulerBatch):
    """Cosine anealing of learning rate (per batch update and not per epoch as in original paper.)
    
    Arguments:
        optimizer {torch.optim.Optimizer} -- Torch optimizer.
    
    Keyword Arguments:
        cycle_len {int} -- Length of cycle (T_i in paper) (default: {1})
        cycle_multiplier {int} -- Multiplier of cycle_len after each finished cycle (default: {2})
        cycle_eta_multiplier {float} -- Multiply eta with this number after each finished cycle.
            Reaonalble values should be between 1 and 0. (default: {1.})
        eta_min {int} -- Minimul learing rate muliplier (will not actually be zero) (default: {0})
        last_batch {int} -- Index of last batch (default: {-1})
    """
    def __init__(self, optimizer, cycle_len=1, cycle_multiplier=2, cycle_eta_multiplier=1.,
                 eta_min=0, batch_iter=-1, keep_etas=True):
        super().__init__(optimizer, batch_iter)
        self.cycle_len = cycle_len
        self.cycle_multiplier = cycle_multiplier
        if (cycle_eta_multiplier > 1) or (cycle_eta_multiplier < 0):
            raise ValueError("cycle_eta_multiplier should be in (0, 1].")
        self.cycle_eta_multiplier = cycle_eta_multiplier
        self.eta_max = 1.
        self.eta_min = eta_min
        self.keep_etas = keep_etas
        if not keep_etas:
            raise ValueError("For now, we require that 'keep_etas' is True")
        self.etas = []
        self.cycle_iter = self.batch_iter % self.cycle_len
    
    @property
    def cycle_len(self):
        return self._cycle_len
    
    @cycle_len.setter
    def cycle_len(self, cycle_len):
        if cycle_len < 1:
            raise ValueError(f"Need cycle_len >= 1.")
        assert type(cycle_len) is int
        self._cycle_len = cycle_len

    def get_lr(self):
        # eta = (self.eta_min + 0.5 * (1. - self.eta_min)
        #        * (1 + math.cos(math.pi * self.cycle_iter / self.cycle_len)))
        eta = (self.eta_min + 0.5 * (self.eta_max - self.eta_min)
               * (1 + math.cos(math.pi * self.cycle_iter / self.cycle_len)))
        if self.keep_etas:
            self.etas.append(eta)
        return [eta * group['initial_lr'] for group in self.optimizer.param_groups]

    def step(self, score=None, batch_iter=None):
        super().step(score, batch_iter)
        self.cycle_iter += 1
        if self.cycle_iter == self.cycle_len:
            self.cycle_iter = 0
            self.cycle_len *= self.cycle_multiplier
            self.eta_max *= self.cycle_eta_multiplier

    def to_pandas(self):
        etas = pd.Series(self.etas).rename('eta')
        etas.index.name = 'batch'
        return etas

    def plot(self, **kwargs):
        return self.to_pandas().to_frame().plot(**kwargs)
